// This code example is created for educational purpose
// by Thorsten Thormaehlen (contact: www.thormae.de).
// It is distributed without any warranty.

package de.unimarburg.objtovboexampleapp;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;

import javax.microedition.khronos.egl.EGLConfig;
import javax.microedition.khronos.opengles.GL10;
import javax.microedition.khronos.opengles.GL11;

import android.app.Activity;
import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Matrix;
import android.opengl.GLSurfaceView;
import android.opengl.GLU;
import android.opengl.GLUtils;
import android.os.Bundle;
import android.os.SystemClock;

class Renderer {
  public float t = 0.0f;
  public int mode = 0;

  private Context appContext;
  private int[] vertBufID = new int[1];
  private int texID = 0;
  private int vertNo = 0;
  private int SIZEOF_FLOAT = Float.SIZE / 8;

  public Renderer(Context context) {
    appContext = context;
  }

  public void init(GL10 gl10, EGLConfig config) {
    GL11 gl = (GL11) gl10;
    gl.glEnable(GL11.GL_DEPTH_TEST);

    // assuming the following structure of input vertex data
    // struct Vertex {
    //   float position[3];
    //   float color[4];
    //   float texCoord[2];
    //   float normal[3];
    // };
  
    int perVertexFloats = (3+4+2+3);
    float vertexData[] = loadVertexData(R.raw.pushbike, perVertexFloats);

    vertNo = vertexData.length / perVertexFloats;

    ByteBuffer bb = ByteBuffer.allocateDirect(vertexData.length * SIZEOF_FLOAT);
    bb.order(ByteOrder.nativeOrder());
    FloatBuffer dataIn = bb.asFloatBuffer();
    dataIn.put(vertexData);
    dataIn.flip();

    // generating vertex VBO
    gl.glGenBuffers(1, vertBufID, 0);
    gl.glBindBuffer(GL11.GL_ARRAY_BUFFER, vertBufID[0]);
    gl.glBufferData(GL11.GL_ARRAY_BUFFER, dataIn.capacity() * SIZEOF_FLOAT,
        dataIn, GL11.GL_STATIC_DRAW);

    texID = loadTexture(gl, R.drawable.checkerboard);
  }

  public void resize(GL10 gl10, int width, int height) {
    GL11 gl = (GL11) gl10;
    gl.glViewport(0, 0, width, height);
    gl.glMatrixMode(GL11.GL_PROJECTION);
    gl.glLoadIdentity();
    GLU.gluPerspective(gl, 30.0f, (float) width / (float) height, 1.0f, 10.0f);
  }

  public void display(GL10 gl10) {
    GL11 gl = (GL11) gl10;
    gl.glClearColor(0.3f, 0.3f, 0.3f, 0.0f);
    gl.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);

    gl.glMatrixMode(GL11.GL_MODELVIEW);
    gl.glLoadIdentity();
    
    // camera orbits in the z=1.5 plane
    // and looks at origin
    double rad = Math.PI / 180.0f * t;
    GLU.gluLookAt(gl, 1.5f*(float)Math.cos(rad), 1.5f*(float)Math.sin(rad), 1.5f, // eye
                  0.0f, 0.0f, 0.0f, // look at
                  0.0f, 0.0f, 1.0f); // up
    
    // activating VBO
    gl.glBindBuffer(GL11.GL_ARRAY_BUFFER, vertBufID[0]);
    int stride = (3 + 4 + 2 + 3) * SIZEOF_FLOAT;
    int offset = 0;

    // position
    gl.glVertexPointer(3, GL11.GL_FLOAT, stride, offset);
    gl.glEnableClientState(GL11.GL_VERTEX_ARRAY);

    // color
    offset = 0 + 3 * SIZEOF_FLOAT;
    gl.glColorPointer(4, GL11.GL_FLOAT, stride, offset);
    gl.glEnableClientState(GL11.GL_COLOR_ARRAY);

    // texture
    offset = 0 + (3 + 4) * SIZEOF_FLOAT;
    gl.glTexCoordPointer(2, GL11.GL_FLOAT, stride, offset);
    gl.glEnableClientState(GL11.GL_TEXTURE_COORD_ARRAY);

    // normals
    offset = 0 + (3 + 4 + 2) * SIZEOF_FLOAT;
    gl.glNormalPointer(GL11.GL_FLOAT, stride, offset);
    gl.glEnableClientState(GL11.GL_NORMAL_ARRAY);

    // bind texture
    if (mode == 0) {
      gl.glEnable(GL11.GL_TEXTURE_2D);
      gl.glBindTexture(GL11.GL_TEXTURE_2D, texID);
    } else {
      gl.glDisable(GL11.GL_TEXTURE_2D);
      gl.glBindTexture(GL11.GL_TEXTURE_2D, 0);
    }

    // render data
    gl.glDrawArrays(GL11.GL_TRIANGLES, 0, vertNo);

    gl.glDisableClientState(GL11.GL_VERTEX_ARRAY);
    gl.glDisableClientState(GL11.GL_COLOR_ARRAY);
    gl.glDisableClientState(GL11.GL_TEXTURE_COORD_ARRAY);
    gl.glDisableClientState(GL11.GL_NORMAL_ARRAY);
    gl.glDisable(GL11.GL_TEXTURE_2D);

    gl.glFlush();
  }

  // returns a valid textureID on success, otherwise 0
  private int loadTexture(GL10 gl10, int resourceId) {
    GL11 gl = (GL11) gl10;

    int level = 0;
    int border = 0;

    // read the image from the resources using BitmapFactory
    BitmapFactory.Options options = new BitmapFactory.Options();
    options.inScaled = false; // No scaling
    Bitmap bitmap = BitmapFactory.decodeResource(appContext.getResources(),
        resourceId, options);

    // we need to vertically flip the image because the image origin
    // in OpenGL is the lower-left corner.
    Matrix flipMat = new Matrix();
    flipMat.setScale(1.0f, -1.0f);
    Bitmap flippedBitmap = Bitmap.createBitmap(bitmap, 0, 0, bitmap.getWidth(),
        bitmap.getHeight(), flipMat, false);
    bitmap.recycle();

    // request textureID
    final int[] textureID = new int[1];
    gl.glGenTextures(1, textureID, 0);
    if (textureID[0] == 0)
      return 0;

    // bind texture
    gl.glBindTexture(GL11.GL_TEXTURE_2D, textureID[0]);

    // define how to filter the texture
    gl.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER,
        GL11.GL_LINEAR);
    gl.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER,
        GL11.GL_LINEAR);

    // texture colors modulate the original color values
    gl.glTexEnvf(GL11.GL_TEXTURE_ENV, GL11.GL_TEXTURE_ENV_MODE,
        GL11.GL_MODULATE);

    // specify the 2D texture map
    GLUtils.texImage2D(GL11.GL_TEXTURE_2D, level, flippedBitmap, border);

    // the texture data is now handled by OpenGL, we can free the local copy
    flippedBitmap.recycle();

    return textureID[0];

  }

  private float[] loadVertexData(int resourceId, int perVertexFloats) {

    float[] floatArray = new float[0];
    // read vertex data from file
    int vertSize = 0;
    try {
      InputStream is = appContext.getResources().openRawResource(resourceId);
      BufferedReader br = new BufferedReader(new InputStreamReader(is));
      String line = br.readLine();
      if (line != null) {
        vertSize = Integer.parseInt(line);
        floatArray = new float[vertSize];
      }
      int i = 0;
      while ((line = br.readLine()) != null && i < floatArray.length) {
        floatArray[i] = Float.parseFloat(line);
        i++;
      }
      if (i != vertSize || (vertSize % perVertexFloats) != 0) {
        floatArray = new float[0];
      }
      br.close();
    } catch (IOException e) {
      e.printStackTrace();
    }
    
    return floatArray;
  }

}

class GLSurfaceViewRenderer implements GLSurfaceView.Renderer {

  private Renderer renderer;

  final int targetFPS = 60;
  final int targetMillis = 1000 / targetFPS;
  private long startTime = 0;

  public GLSurfaceViewRenderer(Context appContext) {
    renderer = new Renderer(appContext);
  }

  @Override
  public void onDrawFrame(GL10 gl10) {

    float offset = 0.5f;
    renderer.t += offset;
    renderer.display(gl10);

    // this code attempts to achieve a target frames-per-second rate
    long timeTillNextRender = startTime + targetMillis;
    long sleepTime = timeTillNextRender - SystemClock.uptimeMillis();
    if (sleepTime > 0) {
      try {
        Thread.sleep(sleepTime);
      } catch (InterruptedException e) {
        e.printStackTrace();
      }
    } else {
      // rendering is too slow
    }
    startTime = SystemClock.uptimeMillis();
  }

  @Override
  public void onSurfaceChanged(GL10 gl10, int width, int height) {
    renderer.resize(gl10, width, height);
  }

  @Override
  public void onSurfaceCreated(GL10 gl10, EGLConfig config) {
    renderer.init(gl10, config);
    startTime = SystemClock.uptimeMillis();
  }

  public Renderer getRenderer() {
    return renderer;
  }
}

public class ObjToVboExample extends Activity {

  private GLSurfaceView myGLSurfaceView;
  private GLSurfaceViewRenderer myGLSurfaceViewRenderer;

  @Override
  public void onCreate(Bundle savedInstanceState) {
    super.onCreate(savedInstanceState);
    myGLSurfaceView = new GLSurfaceView(this);
    myGLSurfaceViewRenderer = new GLSurfaceViewRenderer(this);
    myGLSurfaceView.setRenderer(myGLSurfaceViewRenderer);
    // myGLSurfaceView.setRenderMode(GLSurfaceView.RENDERMODE_WHEN_DIRTY);
    setContentView(myGLSurfaceView);
  }

  @Override
  protected void onPause() {
    // Large objects in graphics memory should be de-allocated here
    // This is skipped here for simplicity of the example.
    super.onPause();
    myGLSurfaceView.onPause();
  }

  @Override
  protected void onResume() {
    super.onResume();
    // Large objects in graphics memory should be re-created here
    // if they were de-allocated on onPause().
    // This is skipped here for simplicity of the example.
    myGLSurfaceView.onResume();
  }
}
